//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.04
// * attempt_move() function added
// * attempt_move() allows null move and keeps track of last position
// * split GameState off into separate .cpp
// * execute_move() added

//----------------------------------------------------------------------------

#include "Agent.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// determine ARENA team membership

bool is_odd_team(string s) 
{ 
  return s.find("odd") != string::npos; 
}

//----------------------------------------------------------------------------

bool is_even_team(string s) 
{ 
  return s.find("even") != string::npos; 
}

//----------------------------------------------------------------------------

bool is_ghost(string s) 
{ 
  return s.find("ghost") != string::npos; 
}

//----------------------------------------------------------------------------

bool is_runner(string s) 
{ 
  return s.find("runner") != string::npos; 
}

//----------------------------------------------------------------------------

bool is_power_pill(string s) 
{ 
  return s == "power pill";
}

//----------------------------------------------------------------------------

bool position_in_list(Point_2D p, vector <Point_2D> *V)
{
  if (V) {
    for (int i = 0; i < V->size(); i++)
      if (p == (*V)[i])
	return true;
  }

  return false;
}

//----------------------------------------------------------------------------

// REFERENCES

// * http://www-nlp.stanford.edu/~grenager/cs121/handouts/proj1.pdf
// * http://www-inst.eecs.berkeley.edu/~cs188/fa07/projects/multiagent/multiagentproject.html

// if w_row, w_col is an unvisited neighbor, update distance and add to queue

void BFS_expand(Point_2D w, Point_2D & v, queue <Point_2D> & Q, BlockMaze *M, vector <Point_2D> *obstacles)
{
  if (!position_in_list(w, obstacles) && !M->vertex[w.row][w.col].is_wall() && M->vertex[w.row][w.col].distance == DIST_INFINITY) {

    M->vertex[w.row][w.col].distance = M->vertex[v.row][v.col].distance + 1;
    M->vertex[w.row][w.col].path = &M->vertex[v.row][v.col];

    Q.push(w);
  }
}

//----------------------------------------------------------------------------

// breadth-first search: mark all block vertices with 
// minimum distance to start location

void BFS(Point_2D start, BlockMaze *M, vector <Point_2D> *obstacles)
{
  queue <Point_2D> Q;
  Point_2D v;

  M->initialize_distances();

  M->vertex[start.row][start.col].distance = 0;
  Q.push(start);
  
  while (!Q.empty()) {

    v = Q.front(); Q.pop();

    if (M->vertex[v.row][v.col].distance != DIST_INFINITY) {

      // N, E, S, W

      BFS_expand(Point_2D(v.row - 1, v.col), v, Q, M, obstacles);
      BFS_expand(Point_2D(v.row, v.col + 1), v, Q, M, obstacles);
      BFS_expand(Point_2D(v.row + 1, v.col), v, Q, M, obstacles);
      BFS_expand(Point_2D(v.row, v.col - 1), v, Q, M, obstacles);

    } 
  }
}

//----------------------------------------------------------------------------

// assuming BFS() has been called and distances set,
// works backward from t to distance 0 vertex (aka the start)
// and fills P with sequence

// returns false if path does not exist

bool compute_BFS_path(Point_2D target, vector <Point_2D> & P, BlockMaze *M)
{
  BlockVertex *next_vertex;

  P.clear();
  
  while (1) {

    P.insert(P.begin(), target);

    // stop when one away from start--we don't need that in path

    if (M->vertex[target.row][target.col].distance == 0)
      break;

    next_vertex = M->vertex[target.row][target.col].path;

    if (next_vertex)
      target = next_vertex->pos;
    else
      return false;
  }

  return true;
}

//----------------------------------------------------------------------------

// assuming BFS has been called from start position already

bool closest_power_pill(int & closest_pill_distance, int & closest_pill_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_pill = false;
  vector <Point_2D> P;

  closest_pill_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_power_pill(GS.object_name[i]) && compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_pill = true;
      if (P.size() < closest_pill_distance) {
	closest_pill_distance = P.size();
	closest_pill_index = i;
      }
    }

  return at_least_one_pill;
}

//----------------------------------------------------------------------------

// how far is nearest power pill that EVEN team wants to eat?
// assuming BFS has been called from start position already

bool closest_even_power_pill(int & closest_pill_distance, int & closest_pill_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_pill = false;
  vector <Point_2D> P;

  closest_pill_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_power_pill(GS.object_name[i]) && M->is_odd_side(GS.object_position[i]) &&
	compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_pill = true;
      if (P.size() < closest_pill_distance) {
	closest_pill_distance = P.size();
	closest_pill_index = i;
      }
    }

  return at_least_one_pill;
}

//----------------------------------------------------------------------------

// how far is nearest power pill that ODD team wants to eat?
// assuming BFS has been called from start position already

bool closest_odd_power_pill(int & closest_pill_distance, int & closest_pill_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_pill = false;
  vector <Point_2D> P;

  closest_pill_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_power_pill(GS.object_name[i]) && M->is_even_side(GS.object_position[i]) &&
	compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_pill = true;
      if (P.size() < closest_pill_distance) {
	closest_pill_distance = P.size();
	closest_pill_index = i;
      }
    }

  return at_least_one_pill;
}

//----------------------------------------------------------------------------

// how far is nearest power pill that EVEN team wants to eat?
// assuming BFS has been called from start position already

bool closest_available_even_power_pill(vector <bool> & taken, int & closest_pill_distance, int & closest_pill_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_pill = false;
  vector <Point_2D> P;

  closest_pill_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_power_pill(GS.object_name[i]) && M->is_odd_side(GS.object_position[i]) && !taken[i] &&
	compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_pill = true;
      if (P.size() < closest_pill_distance) {
	closest_pill_distance = P.size();
	closest_pill_index = i;
      }
    }

  return at_least_one_pill;
}

//----------------------------------------------------------------------------

// how far is nearest power pill that ODD team wants to eat?
// assuming BFS has been called from start position already

bool closest_available_odd_power_pill(vector <bool> & taken, int & closest_pill_distance, int & closest_pill_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_pill = false;
  vector <Point_2D> P;

  closest_pill_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_power_pill(GS.object_name[i]) && M->is_even_side(GS.object_position[i]) && !taken[i] &&
	compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_pill = true;
      if (P.size() < closest_pill_distance) {
	closest_pill_distance = P.size();
	closest_pill_index = i;
      }
    }

  return at_least_one_pill;
}

//----------------------------------------------------------------------------

// assuming BFS has been called from start position already

bool closest_ghost(int & closest_ghost_distance, int & closest_ghost_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_ghost = false;
  vector <Point_2D> P;

  closest_ghost_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_ghost(GS.object_name[i]) && compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_ghost = true;
      if (P.size() < closest_ghost_distance) {
	closest_ghost_distance = P.size();
	closest_ghost_index = i;
      }
    }

  return at_least_one_ghost;
}

//----------------------------------------------------------------------------

// assuming BFS has been called from start position already

bool closest_runner(int & closest_runner_distance, int & closest_runner_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_runner = false;
  vector <Point_2D> P;

  closest_runner_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_runner(GS.object_name[i]) && compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_runner = true;
      if (P.size() < closest_runner_distance) {
	closest_runner_distance = P.size();
	closest_runner_index = i;
      }
    }

  return at_least_one_runner;
}

//----------------------------------------------------------------------------

// assuming BFS has been called from start position already

bool closest_available_even_runner(vector <bool> & taken, int & closest_runner_distance, int & closest_runner_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_runner = false;
  vector <Point_2D> P;

  closest_runner_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_runner(GS.object_name[i]) && is_even_team(GS.object_name[i]) && !taken[i] && 
	compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_runner = true;
      if (P.size() < closest_runner_distance) {
	closest_runner_distance = P.size();
	closest_runner_index = i;
      }
    }

  return at_least_one_runner;
}

//----------------------------------------------------------------------------

// assuming BFS has been called from start position already

bool closest_even_runner(int & closest_runner_distance, int & closest_runner_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_runner = false;
  vector <Point_2D> P;

  closest_runner_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_runner(GS.object_name[i]) && is_even_team(GS.object_name[i]) &&
      compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_runner = true;
      if (P.size() < closest_runner_distance) {
	closest_runner_distance = P.size();
	closest_runner_index = i;
      }
    }

  return at_least_one_runner;
}

//----------------------------------------------------------------------------

// assuming BFS has been called from start position already

bool closest_available_odd_runner(vector <bool> & taken, int & closest_runner_distance, int & closest_runner_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_runner = false;
  vector <Point_2D> P;

  closest_runner_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_runner(GS.object_name[i]) && is_odd_team(GS.object_name[i]) && !taken[i] &&
	compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_runner = true;
      if (P.size() < closest_runner_distance) {
	closest_runner_distance = P.size();
	closest_runner_index = i;
      }
    }

  return at_least_one_runner;
}

//----------------------------------------------------------------------------

// assuming BFS has been called from start position already

bool closest_odd_runner(int & closest_runner_distance, int & closest_runner_index, const GameState & GS, BlockMaze *M)
{
  int i;
  bool at_least_one_runner = false;
  vector <Point_2D> P;

  closest_runner_distance = 100000;

  for (i = 0; i < GS.object_position.size(); i++) 
    if (is_runner(GS.object_name[i]) && is_odd_team(GS.object_name[i]) &&
      compute_BFS_path(GS.object_position[i], P, M)) {
      at_least_one_runner = true;
      if (P.size() < closest_runner_distance) {
	closest_runner_distance = P.size();
	closest_runner_index = i;
      }
    }

  return at_least_one_runner;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

Agent::Agent(string na, char glyph, int color, int initial_x, int initial_y)
{
  name = na;
  initial_position.x = initial_x;
  initial_position.y = initial_y;
  last_position = position = initial_position;
  move_approved = false;

  terrain_wait_counter = 0;

  draw_glyph = glyph;
  draw_color = default_draw_color = color;
  do_draw_path = false;
  do_draw_distances = false;
  do_draw_name = false;

  turn_number = 0;
}

//----------------------------------------------------------------------------

void Agent::reset()
{
  last_position = position = initial_position;
  move_approved = false;

  terrain_wait_counter = 0;

  draw_color = default_draw_color;

  do_draw_path = false;
  do_draw_distances = false;
  do_draw_name = false;

  turn_number = 0;

  path.clear();
}

//----------------------------------------------------------------------------

// this is the only way an agent should move--DO NOT SET POSITION DIRECTLY!!!
// also, there should only be one call to attempt_move() per call to update().

bool Agent::attempt_move(int row, int col, BlockMaze *M)
{
  move_approved = false;

  // is move legal (to interior of maze, no diagonals, no hopping or teleporting)?
  
  if (row <= 0 || row >= M->num_rows - 1 || col <= 0 || col >= M->num_cols - 1)
    goto no_move;

  if (!((col == position.col && row == position.row) ||         // don't move at all
	(col == position.col + 1 && row == position.row) ||     // move E
	(col == position.col - 1 && row == position.row) ||     // move W
	(col == position.col && row == position.row + 1) ||     // move S
	(col == position.col && row == position.row - 1)))      // move N
    goto no_move;

  // is move to an empty space?

  if (M->vertex[row][col].is_wall())
    goto no_move;

  // does terrain allow it right now (i.e, agent may have to "wait" a few turns)

  if (terrain_wait_counter > 0) 
    goto no_move;

  // move is approved...

  // ...if destination is terrain, set wait counter

  if (M->vertex[row][col].is_water())
    terrain_wait_counter = WATER_COST;
  else if (M->vertex[row][col].is_fire())
    terrain_wait_counter = FIRE_COST;
  else if (M->vertex[row][col].is_wind() && is_ghost(name))
    terrain_wait_counter = WIND_COST;

  // ...ok---actually do it now!

  move_position.col = col;
  move_position.row = row;

  move_approved = true;

  return true;

 no_move: 
  move_position = position;
  move_approved = true;

  return false;
}

//----------------------------------------------------------------------------

void Agent::execute_move(BlockMaze *M)
{
  if (move_approved) {
    last_position = position;
    position = move_position;
  }
}

//----------------------------------------------------------------------------

// react to keyboard input

void Agent::handle_key(char c, BlockMaze *M)
{
  switch (c) {
    
    // move up	
      
  case NORTH_KEY:
    attempt_move(position.y - 1, position.x, M);
    break;
    
    // move down
    
  case SOUTH_KEY:
    attempt_move(position.y + 1, position.x, M);
    break;

    // move left

  case WEST_KEY:
    attempt_move(position.y, position.x - 1, M);
    break;

    // move right

  case EAST_KEY:
    attempt_move(position.y, position.x + 1, M);
    break;

    // change various drawing debug modes

  case DRAW_DISTANCE_KEY:
    do_draw_distances = !do_draw_distances;
    break;

  case DRAW_PATH_KEY:
    do_draw_path = !do_draw_path;
    break;

  case DRAW_NAME_KEY:
    do_draw_name = !do_draw_name;
    break;
  }

}

//----------------------------------------------------------------------------

// draw optional underlay of distances/path/name, etc. to
// help with debugging or to clarify agent's internal state

void Agent::draw_debug(BlockMaze *M)
{
  // BFS output -- note that distances in M "belong" to last
  // agent to run BFS and not necessarily this one

  if (do_draw_distances)
    M->draw_distances();

  // path to target in agent color

  attron(COLOR_PAIR(draw_color));  

  if (do_draw_path) {
    for (int i = 0; i < path.size(); i++) {
      move(path[i].y, path[i].x);
      printw("*");
    }
  }

  attroff(COLOR_PAIR(draw_color));  

  // name?

  if (do_draw_name) {
    move(position.y, position.x + 1);
    printw("%s", name.c_str());
  }
}

//----------------------------------------------------------------------------

// draw the agent itself--always on top of debug information, even from
// other agents

void Agent::draw(BlockMaze *M)
{
  attron(COLOR_PAIR(draw_color));  

  move(position.y, position.x);
  printw("%c", draw_glyph);

  attroff(COLOR_PAIR(draw_color));  
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

