//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.07
// * added score/time elapsed drawing
// * crossover collisions now detected
// * added wind terrain
// * added blue_ghost_wait_counter
// * starting to move functions, variables over to GameState class; making it const in call to update()
// * added "-blue" option for blue ghosts/no blue ghosts; saving .lay; new quit key
// * moved to MazeRunner.cpp
// * added arena_mode flag

// sample arena rules (not exactly ours):
// http://inst.eecs.berkeley.edu/~cs188/fa10/projects/contest/contest.html

//----------------------------------------------------------------------------

#include "MazeRunner.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// random maze constructor

MazeRunner::MazeRunner(int rows, int cols, int bg_time_limit, int mt) 
{ 
  num_updates = 0;
  GS.blue_ghost_wait_counter = 0;
  blue_ghost_time_limit = bg_time_limit; 
  arena_mode = false;
  arena_border_col = cols / 2;
  do_step = false;
  max_turns = mt;

  M = new BlockMaze(rows, cols);   

  // loop until we get a good random maze...haven't seen a failure
  // yet at this level

  int i;
  for (i = 0; i < 5; i++)
    if (M->knock_down_walls_randomly_until_everything_connected())
      break;
  if (i == 5) {
    cout << "Sorry but I failed to generate a maze...your dimensions might be too big\n";
    exit(1);
  }

  // define this in header file if you want random terrain cells added
#ifdef DO_TERRAIN
  M->add_random_terrain(0.1, 0.1);
#endif

  // place runner

  runner_agent_index = agent.size();
  Agent *runner = new RunnerAgent("runner", 'c', YELLOW_COLOR);
  M->pick_random_empty_space(runner->position.y, runner->position.x);
  agent.push_back(runner);

  // place ghosts

  // ???

  // place power pills

  // at least 1 power pill, plus one more for every 100 squares. 
  // make sure they have some minimal separation and are not on top
  // of runner to start.  not sure if this is impossible for some
  // configurations--haven't seen it yet...
  
  int area = rows * cols;
  int num_power_pills = 1 + area / 100;
  Point_2D p;
  int min_dist, dist;

  for (int i = 0; i < num_power_pills; i++) {
    while (1) {
      M->pick_random_empty_space(p.y, p.x);
      if (p.x == agent[runner_agent_index]->position.x && p.y == agent[runner_agent_index]->position.y)
	continue;
      min_dist = 100000;
      for (int j = 0; j < power_pill.size(); j++) {
	dist = fabs(p.x - power_pill[j].x) + fabs(p.y - power_pill[j].y);
	if (dist < min_dist)
	  min_dist = dist;
      }
      if (min_dist >= 5)
	break;
    }
    power_pill.push_back(p);
  }

  initial_power_pill = power_pill;
  GS.initial_num_power_pills = initial_power_pill.size();
}

//----------------------------------------------------------------------------

// load from .lay constructor

MazeRunner::MazeRunner(string filename, int bg_time_limit, int mt, bool self_play)
{
  num_updates = 0;
  GS.blue_ghost_wait_counter = 0;
  blue_ghost_time_limit = bg_time_limit; 
  arena_mode = false;
  do_step = false;
  max_turns = mt;

  M = new BlockMaze(MIN_MAZE_ROWS, MIN_MAZE_COLS);
  
  load_lay(filename, self_play);
  arena_border_col = M->num_cols / 2;
  M->compute_wall_counts();
  initial_power_pill = power_pill;
  GS.initial_num_power_pills = initial_power_pill.size();

  // don't change terrain
}

//----------------------------------------------------------------------------
// ARENA functions
//----------------------------------------------------------------------------

// how many pills are left for the ODD team to eat?

int MazeRunner::num_odd_pills(vector <Point_2D> & pill)
{
  int i, total;

  for (i = total = 0; i < pill.size(); i++)
    if (M->is_even_side(pill[i]))
      total++;

  return total;
}

//----------------------------------------------------------------------------

// how many pills are left for the EVEN team to eat?

int MazeRunner::num_even_pills(vector <Point_2D> & pill)
{
  int i, total;

  for (i = total = 0; i < pill.size(); i++)
    if (M->is_odd_side(pill[i]))
      total++;

  return total;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// render current state of everything

void MazeRunner::draw()
{
  int i;

  // clear screen

  erase();

  // draw maze itself

  M->draw();

  // draw ARENA border line

  if (arena_mode) {
    for (i = 0; i < M->num_rows; i++) {
      if (M->vertex[i][M->num_cols / 2].is_empty()) {
	move(i, arena_border_col);
	printw("|");
      }
    }
  }

  // "score" info

  move(M->num_rows, 0);
  printw("%i turns elapsed", num_updates);

  if (!arena_mode) {
    move(M->num_rows + 1, 0);
    printw("%i pills to go", power_pill.size());
  }
  else {
    move(M->num_rows + 1, 0);
    printw("Odd: %i/%i remain", num_odd_pills(power_pill), num_odd_pills(initial_power_pill));

    move(M->num_rows + 2, 0);
    printw("Even: %i/%i remain", num_even_pills(power_pill), num_even_pills(initial_power_pill));
  }

  // draw agents, power pills

  for (i = 0; i < agent.size(); i++) 
    agent[i]->draw_debug(M);

  draw_power_pills();

  for (i = 0; i < agent.size(); i++) {

    if (!arena_mode) {

      if (is_ghost(agent[i]->name)) {
	if (GS.blue_ghost_wait_counter > 0)
	  agent[i]->draw_color = BLUE_COLOR;
	else
	  agent[i]->draw_color = agent[i]->default_draw_color;
      }
    }
    else {
      if ((is_odd_team(agent[i]->name) && M->is_even_side(agent[i]->position)) ||
	  (is_even_team(agent[i]->name) && M->is_odd_side(agent[i]->position)))
	agent[i]->draw_color = BLUE_COLOR;
      else
	agent[i]->draw_color = agent[i]->default_draw_color;
    }

    agent[i]->draw(M);
  }

  // make drawing visible

  refresh();
}

//----------------------------------------------------------------------------

// write current state of everything to Berkeley .lay file

// does NOT write terrain information, since default random mazes
// don't have terrain in them

void MazeRunner::save_lay(string filename)
{
  ofstream outStream;
  char c;
  int row, col;

  // try to open file to mutate

  outStream.open(filename.c_str());

  if (outStream.fail()) {
    cout << "Failed to open file " << filename << endl;
    exit(1);
  }

  // write maze and everything in it

  for (row = 0; row < M->num_rows; row++) {
    for (col = 0; col < M->num_cols; col++) {

      if (M->is_wall(row, col))
	outStream.put(LAY_WALL_CHAR); 
      else if (GS.object_here("power pill", row, col))
	outStream.put(LAY_POWER_PILL_CHAR);
      else if (GS.object_here("runner", row, col))
	outStream.put(LAY_RUNNER_CHAR);
      else if (GS.object_here("clyde", row, col))
	outStream.put(LAY_CLYDE_CHAR);
      else if (GS.object_here("inky", row, col))
	outStream.put(LAY_INKY_CHAR);
      else if (GS.object_here("blinky", row, col))
	outStream.put(LAY_BLINKY_CHAR);
      else
	outStream.put(' ');
    }
    outStream.put('\n');
  }

  outStream.close();
}

//----------------------------------------------------------------------------

void MazeRunner::draw_power_pills()
{
  int i;

  if (!arena_mode) {
    
    attron(COLOR_PAIR(WHITE_COLOR));  

    for (i = 0; i < power_pill.size(); i++) {
      move(power_pill[i].y, power_pill[i].x);
      printw("o");
    }

    attroff(COLOR_PAIR(WHITE_COLOR));  
  }
  else {

    for (i = 0; i < power_pill.size(); i++) {

      if (M->is_even_side(power_pill[i]))
	attron(COLOR_PAIR(ODD_COLOR));  
      else
	attron(COLOR_PAIR(EVEN_COLOR));  

      move(power_pill[i].y, power_pill[i].x);
      printw("o");

      if (M->is_even_side(power_pill[i]))
	attroff(COLOR_PAIR(ODD_COLOR));  
      else
	attroff(COLOR_PAIR(EVEN_COLOR));  

    }

  }
}

//----------------------------------------------------------------------------

// did a runner consume a power pill?

bool MazeRunner::check_for_power_pill_eaten(int & which_pill)
{
  int i, j;

  for (i = 0; i < power_pill.size(); i++)
    if (power_pill[i] == agent[runner_agent_index]->position) {
      which_pill = i;
      return true;
    }

  return false;
}

//----------------------------------------------------------------------------

void MazeRunner::reset()
{
  int i;

  num_updates = 0;
  GS.blue_ghost_wait_counter = 0;

  for (i = 0; i < agent.size(); i++)
    agent[i]->reset();

  power_pill = initial_power_pill;
}

//----------------------------------------------------------------------------

// do computations, etc. for ARENA

int MazeRunner::arena_update()
{
  int i, which_pill, which_agent;
  bool win_for_odd, win_for_even;

  num_updates++;

  // pack up BEFORE game state

  GS.update(agent, power_pill);

  // let agents do their thing  -- thinking and movement separate for simultaneity

  for (i = 0; i < agent.size(); i++) 
    agent[i]->update(GS, M);

  for (i = 0; i < agent.size(); i++) 
    agent[i]->execute_move(M);
  
  // get rid of all just-eaten power pills

  while (1) {
    if (arena_check_for_power_pill_eaten(which_pill)) 
      power_pill.erase(power_pill.begin() + which_pill);
    else
      break;
  }

  // pack up AFTER game state

  GS.update(agent, power_pill);

  // decrement non-zero wait counter, whether or not agent tried to move
  
  for (i = 0; i < agent.size(); i++) 
    if (agent[i]->terrain_wait_counter > 0)
      agent[i]->terrain_wait_counter--;

  // all done?

  win_for_odd = win_for_even = false;

  if (GS.arena_check_for_odd_level_cleared(M))
    win_for_odd = true;
  
  if (GS.arena_check_for_even_level_cleared(M))
    win_for_even = true;
  if (win_for_odd) {
    if (win_for_even)
      return EVEN_ODD_TIE;
    else
      return ODD_WIN;
  }
  else if (win_for_even)
    return EVEN_WIN;
  
  // if you get caught on opposite side, you get sent back to the start :(

  while (1) {
    if (GS.arena_check_for_runner_caught(which_agent, M)) {
      agent[which_agent]->reset();
      GS.update(agent, power_pill);
    }
    else
      break;
  }

  // time's up

  if (num_updates >= max_turns) {
    int num_odd = num_odd_pills(power_pill);
    int num_even = num_even_pills(power_pill);
    if (num_odd < num_even)
      return ODD_WIN;
    else if (num_even < num_odd)
      return EVEN_WIN;
    else
      return EVEN_ODD_TIE;
  }

  return CONTINUE_PLAYING;
}

//----------------------------------------------------------------------------

// did a runner consume a power pill on the correct side of the arena?
// this only detects one runner on top of one power pill, so it must be
// called multiple times

bool MazeRunner::arena_check_for_power_pill_eaten(int & which_pill)
{
  int i, j;

  for (i = 0; i < power_pill.size(); i++)
    for (j = 0; j < agent.size(); j++) {
      if (agent[j]->position == power_pill[i] &&
	  ((is_odd_team(agent[j]->name) && M->is_even_side(power_pill[i])) ||
	   (is_even_team(agent[j]->name) && M->is_odd_side(power_pill[i])))) {
	which_pill = i;
	return true;
      }
    }

  return false;
}

//----------------------------------------------------------------------------

// do computations, etc. for PAC-MAN

int MazeRunner::update()
{
  int i, which_ghost, which_pill;

  num_updates++;

  // pack up BEFORE game state

  GS.update(agent, power_pill);

  // let agents do their thing

  for (i = 0; i < agent.size(); i++) 
    agent[i]->update(GS, M);

  for (i = 0; i < agent.size(); i++) 
    agent[i]->execute_move(M);

  // pack up AFTER game state

  GS.update(agent, power_pill);

  // decrement non-zero wait counter, whether or not agent tried to move
  
  for (i = 0; i < agent.size(); i++) 
    if (agent[i]->terrain_wait_counter > 0)
      agent[i]->terrain_wait_counter--;

  // consequences?

  if (GS.check_for_runner_caught(which_ghost)) {

    // ghost is blue...send it back to where it started and make it stay there awhile

    if (GS.blue_ghost_wait_counter > 0) {
      agent[which_ghost]->position = agent[which_ghost]->last_position = agent[which_ghost]->initial_position; 
      agent[which_ghost]->path.clear();
      agent[which_ghost]->terrain_wait_counter = GS.blue_ghost_wait_counter;
    }

    // ghost is "live" and runner is dead

    else
      return GHOST_WIN;
  }

  if (GS.check_for_level_cleared())
    return RUNNER_WIN;

  // this may be in the wrong place--put before GS.update() above?

  if (check_for_power_pill_eaten(which_pill)) {
    power_pill.erase(power_pill.begin() + which_pill);
    GS.blue_ghost_wait_counter = blue_ghost_time_limit;
  }
  else if (GS.blue_ghost_wait_counter > 0)
    GS.blue_ghost_wait_counter--;

  if (num_updates >= max_turns) 
    return RUNNER_WIN;

  return CONTINUE_PLAYING;
}

//----------------------------------------------------------------------------

// count rows, cols for Berkeley layout file

void MazeRunner::compute_dimensions_lay(ifstream & inStream, int & rows, int & cols)
{
  char c;
  bool got_cols;
  int this_cols;

  rows = cols = 0;
  got_cols = false;
  this_cols = 0;

  while (!inStream.eof()) {

    inStream.get(c);

    if (inStream.eof()) {
      if (got_cols && this_cols > cols / 2)
	rows++;
      break;
    }

    if (c == '\n') {
      rows++;
      got_cols = true;
      this_cols = 0;
    }
    else {
      this_cols++;
      if (!got_cols)
	cols++;
    }

    // echo to console
    
    //    cout.put(c);  
  }

  //  cout << endl;

  // rewind
  
  inStream.clear();
  inStream.seekg(0,fstream::beg);
}

//----------------------------------------------------------------------------

// read maze from Berkeley layout file

void MazeRunner::load_lay(string filename, bool self_play)
{
  ifstream inStream;
  char c;
  int rows, cols;

  // try to open file to mutate

  inStream.open(filename.c_str());

  if (inStream.fail()) {
    cout << "Failed to open file " << filename << endl;
    exit(1);
  }

  // how big is grid?

  compute_dimensions_lay(inStream, rows, cols);

  // initialize all walls 

  M->num_rows = rows;
  M->num_cols = cols;

  M->vertex.resize(M->num_rows);

  for (rows = 0; rows < M->num_rows; rows++) {
    M->vertex[rows].resize(M->num_cols);
    for (cols = 0; cols < M->num_cols; cols++)
      M->vertex[rows][cols].set_position(rows, cols);
  }

  // read 'em in now that we know dimensions

  rows = cols = 0;

  int runner_num = 0;

  while (!inStream.eof()) {

    inStream.get(c);

    if (inStream.eof())
      break;

    if (c == '\n') {
      rows++;
      cols = 0;
    }
    else {

      stringstream out;
      string s;
      Agent *A;
      int value;

      if (c == LAY_WALL_CHAR)
	M->vertex[rows][cols].make_wall();
      else { 

	M->vertex[rows][cols].make_empty();

	// digits mean this is ARENA

	if (isdigit(c)) {
	  value = atoi(&c);
	  out << c; 
	  if (value % 2) {
	    s = "runner " + out.str() + " odd";
	    if (self_play)
	      A = new ArenaAgent(s, c, ODD_COLOR, cols, rows);
	    else
	      //  A = new TeamXAgent(s, c, ODD_COLOR, cols, rows);
	      //A = new ArenaAgent(s, c, ODD_COLOR, cols, rows);
	      //A = new TeamRavenousZombies(s, c, ODD_COLOR, cols, rows);
	    //A = new TeamTardis(s, c, ODD_COLOR, cols, rows);
	    //A = new TeamGoingSolo(s, c, ODD_COLOR, cols, rows);
	      //A = new TeamNoHonorsRequired(s, c, ODD_COLOR, cols, rows);
	      A = new TeamWazzle(s, c, ODD_COLOR, cols, rows);
	  }
	  else {
	    s = "runner " + out.str() + " even";
	    //A = new TeamRavenousZombies(s, c, EVEN_COLOR, cols, rows);
	    A = new TeamGoingSolo(s, c, EVEN_COLOR, cols, rows);
	    //A = new TeamNoHonorsRequired(s, c, EVEN_COLOR, cols, rows);
	    //A = new TeamWazzle(s, c, EVEN_COLOR, cols, rows);
	    //	    A = new TeamTardis(s, c, EVEN_COLOR, cols, rows);
	    //A = new ArenaAgent(s, c, EVEN_COLOR, cols, rows);
	  }
	  agent.push_back(A);
	  arena_mode = true;
	}
	else if (c == LAY_RUNNER_CHAR) {
	  runner_agent_index = agent.size();
	  A = new RunnerAgent("runner", 'c', YELLOW_COLOR, cols, rows);
	  agent.push_back(A);
	}
	else if (c == LAY_GHOST_CHAR || c == LAY_CLYDE_CHAR) {     
	  out << agent.size();
	  s = "ghost clyde " + out.str();
	  A = new ClydeGhostAgent(s, 'A', YELLOW_COLOR, cols, rows);
	  agent.push_back(A);
	}
	else if (c == LAY_INKY_CHAR) {     
	  out << agent.size();
	  s = "ghost inky " + out.str();
	  A = new InkyGhostAgent(s, 'A', GREEN_COLOR, cols, rows);
	  agent.push_back(A);
	}
	else if (c == LAY_BLINKY_CHAR) {     
	  out << agent.size();
	  s = "ghost blinky " + out.str();
	  A = new BlinkyGhostAgent(s, 'A', RED_COLOR, cols, rows);
	  agent.push_back(A);
	}
	else if (c == LAY_POWER_PILL_CHAR)    
	  power_pill.push_back(Point_2D(rows, cols));
	else if (c == LAY_WATER_CHAR)    
	  M->vertex[rows][cols].terrain_type = TERRAIN_WATER;
	else if (c == LAY_FIRE_CHAR)    
	  M->vertex[rows][cols].terrain_type = TERRAIN_FIRE;
	else if (c == LAY_WIND_CHAR)
	  M->vertex[rows][cols].terrain_type = TERRAIN_WIND;
      }

      cols++;
    }
  }

  // clean up

  inStream.close();
}

//----------------------------------------------------------------------------

// interpret keyboard input...return false if user wants to quit

bool MazeRunner::handle_key()
{
  int c;
  stringstream ss;

  c = getch();

  // exit game

  if (c == QUIT_KEY)
    return false;

  // save game

  else if (c == SAVE_KEY) {
    ss << "saved_" << num_updates << ".lay";
    save_lay(ss.str());
  }

  // change control to next runner for ARENA

  else if (arena_mode && c == NEXT_RUNNER_KEY) 
    runner_agent_index = (runner_agent_index + 1) % agent.size();

  // let all agents show names

  else if (c == DRAW_NAME_KEY)
    for (int i = 0; i < agent.size(); i++)
      agent[i]->handle_key(c, M);

  // send all other commands to runner (current one for ARENA)

  else
    agent[runner_agent_index]->handle_key(c, M);
    
  return true;
}

//----------------------------------------------------------------------------

// set up drawing, print instructions, and go

void MazeRunner::draw_splash()
{
  // get ready to draw

  init_ncurses();

#ifdef DO_SPLASH

  // instructions

  printw("Welcome to MazeRunner, the wacky game of energy collection!!\n");
  printw("If you are playing manually, '%c' = up, '%c' = down, '%c' = left, and '%c' = right\n", NORTH_KEY, SOUTH_KEY, WEST_KEY, EAST_KEY);
  printw("Hit '%c' to exit at any time and '%c' to save\n", QUIT_KEY, SAVE_KEY);   
  printw("Ready?\n");   

  refresh();

  // wait on input

  getch();			                                              

#endif

  // need this for animation without user input

  // comment this out to require hitting a key to "step" through game
  if (!do_step)
    timeout(100);
}

//----------------------------------------------------------------------------

// last words...

void MazeRunner::draw_gameover(int result)
{
#ifdef DO_SPLASH
  int c;

  timeout(-1);

  //  erase();

  if (!arena_mode) {

    move(M->num_rows + 2, 0);
    
    printw("Goodbye!!\n");

    if (result == GHOST_WIN) {
      printw("You sure were easy to catch...\n");
      printw("Hit '%c' to exit\n", QUIT_KEY);
    }
    else if (result == RUNNER_WIN) {
      printw("Nice running...\n");
      printw("Hit '%c' to exit\n", QUIT_KEY);
    }   
    else
      printw("Hit any key to exit\n");   
  }

  // ARENA outcomes

  else {

    move(M->num_rows + 3, 0);

    if (result == EVEN_WIN) {
      printw("Team EVEN wins...\n");
      printw("Hit '%c' to exit\n", QUIT_KEY);
    }   
    else if (result == ODD_WIN) {
      printw("Team ODD wins...\n");
      printw("Hit '%c' to exit\n", QUIT_KEY);
    } 
    else if (result == EVEN_ODD_TIE) {
      printw("It's a tie...\n");
      printw("Hit '%c' to exit\n", QUIT_KEY);
    } 
    else
      printw("Hit any key to exit\n");   
  }
  
  refresh();

  if (result == GHOST_WIN || result == RUNNER_WIN) {
    do { c = getch(); } while (c != QUIT_KEY);
  }

  // any key will do

  else
    getch();			                                              

#endif

  // clean up

  terminate_ncurses();
}

//----------------------------------------------------------------------------

// initialize, loop until done, and clean up

int MazeRunner::main_loop(bool do_draw)
{
  int game_result;
  bool key_result;

  // initialize

  if (do_draw)
    draw_splash();

  // main loop

  key_result = true;

  do {

    if (arena_mode)
      game_result = arena_update();
    else
      game_result = update();

    if (do_draw) {
      draw();
      key_result = handle_key();
    }

  } while (game_result == CONTINUE_PLAYING && key_result);

  // clean up

  if (do_draw)
    draw_gameover(game_result);

  return game_result;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
